/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#include "terrain.h"
#include "random.h"
#include <math.h>
#include <GL/glut.h>
#include <vector>
#include <fstream>

Terrain::Terrain(unsigned int width, unsigned int height, float xScale, float yScale, float zScale)
: Matrix<TerrainValue>(width, height) {
	terrainScale.x = xScale;
	terrainScale.y = yScale;
	terrainScale.z = zScale;
	clear(0);
}

void Terrain::clear(float zero) {
	unsigned int terrainWidth = width();
	unsigned int terrainHeight = height();
	unsigned int x;
	unsigned int y;
	vertex normal(0, 1, 0);

	for (x = 0; x < terrainWidth; x++)
		for (y = 0; y < terrainHeight; y++)
		{
			at(x, y).height = zero;
			at(x, y).normal = normal;
		}
}

void Terrain::load(const char *filename, int width, int height, float floor, float ceiling) {
	std::vector<unsigned char> data;
	std::ifstream file(filename, std::ios::binary|std::ios::in);
	data.resize(width*height);
	resize(width, height);

	if (file.read((char *)&*data.begin(), width*height)) {
		for (int x = 0; x < width; x++)
			for (int y = 0; y < height; y++)
				at(x, y).height = float(data[x * width + y]);
	}
	
	float max = at(0, 0).height;
	float min = at(0, 0).height;

	for (int x = 0; x < width; x++)
		for (int y = 0; y < height; y++)
		{
			if (at(x, y).height > max)
				max = at(x, y).height;
			if (at(x, y).height < min)
				min = at(x, y).height;
		}

	for (int x = 0; x < width; x++)
		for (int y = 0; y < height; y++)
			at(x, y).height = (at(x, y).height - min) / (max - min) * (ceiling - floor) + floor;

	generateNormals(0, 0, width - 1, height - 1);
}

void Terrain::generate(int iterations, float floor, float ceiling) {
	Random random(22); /* Seed the random number generator */
	unsigned int terrainWidth = width();
	unsigned int terrainHeight = height();
	
	if (!terrainWidth || !terrainHeight) return;

	
	while (iterations--)
	{
		int s = static_cast<int>((terrainWidth > terrainHeight) ? terrainHeight : terrainWidth);
		s /= 5;
		int x = random.number(0, static_cast<int>(terrainWidth));
		int y = random.number(0, static_cast<int>(terrainHeight));
		int r = random.number(2, s);
		float h = float(random.number(0, 101)) / 100.0f;

		int p;
		int q;
		int i;
		int j;
		for (p = -r; p < r; p++)
		{
			for (q = -r; q < r; q++)
			{
				i = x + p;
				j = y + q;
				if (i >= 0 && j >= 0 && i < int(terrainWidth) && j < int(terrainWidth) && (p * p + q * q <= r * r))
				{
					at(i, j).height += h * (float(r * r) - float(p * p) - float(q * q));
				}
			}
		}
	}

	float max = at(0, 0).height;
	float min = at(0, 0).height;

	for (unsigned int x = 0; x < terrainWidth; x++)
		for (unsigned int y = 0; y < terrainHeight; y++)
		{
			at(x, y).height *= at(x, y).height;
			if (at(x, y).height > max)
				max = at(x, y).height;
			if (at(x, y).height < min)
				min = at(x, y).height;
		}

	for (unsigned int x = 0; x < terrainWidth; x++)
		for (unsigned int y = 0; y < terrainHeight; y++)
			at(x, y).height = (at(x, y).height - min) / (max - min) * (ceiling - floor) + floor;

	generateNormals(0, 0, terrainWidth - 1, terrainHeight - 1);
}

void Terrain::generateNormals(unsigned int left, unsigned int bottom, unsigned int right, unsigned int top) {
	Matrix<vertex> normalOne(right - left + 1, top - bottom + 1);
	Matrix<vertex> normalTwo(right - left + 1, top - bottom + 1);
	unsigned int x;
	unsigned int y;

	for (x = 0; x < (right - left); x++)
	{
		for (y = 0; y < (top - bottom); y++)
		{
			normalOne.at(x, y) = surfaceNormal(x + left, y + bottom, true);
			normalTwo.at(x, y) = surfaceNormal(x + left, y + bottom, false);
		}
	}

	for (x = 0; x < right - left; x++)
	{
		for (y = 0; y < top - bottom; y++)
		{
			vertex sum(0, 0, 0);
			int count = 0;
			if (x >= 0 && y >= 0 && x < right - left + 1 && y < top - bottom + 1)
			{
				sum += normalOne.at(x, y) + normalTwo.at(x, y);
				count += 2;
			}
			if (x - 1 >= 0 && y - 1 >= 0 && x - 1 < right - left + 1 && y - 1 < top - bottom + 1)
			{
				 sum += normalOne.at(x - 1, y - 1) + normalTwo.at(x - 1, y - 1);
				 count += 2;
			}
			if (x - 1 >= 0 && y >= 0 && x - 1 < right - left + 1 && y < top - bottom + 1)
			{
				sum += normalTwo.at(x - 1, y);
				count++;
			}
			if (x >= 0 && y - 1 >= 0 && x < right - left + 1 && y - 1 < top - bottom + 1)
			{
				sum += normalOne.at(x, y - 1);
				count++;
			}
			if (count)
				sum /= static_cast<float>(count);
			else
				std::cout << "No count" << std::endl;
			at(x + left, y + bottom).normal = normalize(sum);
		}
	}
}

vertex Terrain::surfaceNormal(unsigned int x, unsigned int y, bool one) {
	if (one)
	{
		vertex a(terrainScale.x / width(),
			terrainScale.y * (at(x + 1, y + 1).height - at(x, y + 1).height),
			0.0f);
		vertex b(0.0f,
			at(x, y).height - at(x, y + 1).height,
			-terrainScale.z / height());
		return normalize(crossProduct(a, b));
	}
	else
	{
		vertex c(-terrainScale.x / width(),
			terrainScale.y * (at(x, y).height - at(x + 1, y).height),
			0.0f);
		vertex d(0.0f,
			terrainScale.y * (at(x + 1, y + 1).height - at(x + 1, y).height),
			terrainScale.z / height());
		return normalize(crossProduct(c, d));
	}
}

void Terrain::scale(float x, float y, float z) {
	terrainScale.x = x;
	terrainScale.y = y;
	terrainScale.z = z;
	generateNormals(0, 0, width() - 1, height() - 1);
	generateNormals(0, 0, width() - 1, height() - 1);
}

vertex Terrain::scale() {
	return terrainScale;
}

float Terrain::maxHeight(unsigned int x, unsigned int y) {
	float m = 0.0f; // at(x, y).height;

	for (unsigned int i = x - 1; i < x + 1; i++)
	{
		for (unsigned int j = y - 1; j < y + 1; j++)
		{
			if (x >= 0 && x < width() && y >= 0 && y < height() && at(i, j).height > m)
			{
				m = at(i, j).height;
			}
		}
	}
	return m * terrainScale.y;
}

float Terrain::minHeight(unsigned int x, unsigned int y) {
	float m = 0.0f; // at(x, y).height;

	for (unsigned int i = x - 1; i < x + 1; i++)
	{
		for (unsigned int j = y - 1; j < y + 1; j++)
		{
			if (x >= 0 && x < width() && y >= 0 && y < height() && at(i, j).height < m)
			{
				m = at(i, j).height;
			}
		}
	}
	return m * terrainScale.y;
}

void Terrain::sphereDeform(const vertex &position, float radius) {
	int sx1 = static_cast<int>((position.x - radius) / terrainScale.x * width() - 1.0f);
	int sx2 = static_cast<int>((position.x + radius) / terrainScale.x * width() + 1.0f);
	int sy1 = static_cast<int>((position.z - radius) / terrainScale.z * height() - 1.0f);
	int sy2 = static_cast<int>((position.z + radius) / terrainScale.z * height() + 1.0f);

	if (sx1 < 0) sx1 = 0; if (sx1 >= static_cast<int>(width())) sx1 = static_cast<int>(width() - 1);
	if (sx2 < 0) sx2 = 0; if (sx2 >= static_cast<int>(width())) sx2 = static_cast<int>(width() - 1);
	if (sy1 < 0) sy1 = 0; if (sy1 >= static_cast<int>(height())) sy1 = static_cast<int>(height() - 1);
	if (sy2 < 0) sy2 = 0; if (sy2 >= static_cast<int>(height())) sy2 = static_cast<int>(height() - 1);

	unsigned int x1 = static_cast<unsigned int>(sx1);
	unsigned int x2 = static_cast<unsigned int>(sx2);
	unsigned int y1 = static_cast<unsigned int>(sy1);
	unsigned int y2 = static_cast<unsigned int>(sy2);

	unsigned int x;
	unsigned int y;

	for (x = x1; x <= x2; x++)
	{
		for (y = y1; y <= y2; y++)
		{
			float fx = -position.x + x * terrainScale.x / width();
			float fy = -position.z + y * terrainScale.z / height();
			float h = at(x, y).height * terrainScale.y;
			float f = radius * radius - fx * fx - fy * fy;
			if (f > 0)
			{
				float dh = static_cast<float>(sqrt(f));

				if (h > position.y + dh) at(x, y).height += (h - (position.y + dh)) / terrainScale.y;
				if (h > position.y - dh) at(x, y).height -= (h - (position.y - dh)) / terrainScale.y;
				if (at(x, y).height < 0) at(x, y).height = 0.0f;
			}
		}
	}

	generateNormals(0, 0, width() - 1, height() - 1); //x1, y1, x2, y2);
}

void Terrain::display() {
	unsigned int terrainWidth = width();
	unsigned int terrainHeight = height();
	unsigned int x;
	unsigned int y;
	vertex scale = terrainScale;

	/* This is cutting it short, but I had normal problems at the edges and this is the easiest solution */
	for (x = 1; x < terrainWidth - 2; x++)
	{
		for (y = 1; y < terrainHeight - 2; y++)
		{
			TerrainValue tv[2][2];
			tv[0][0] = at(x, y);
			tv[0][1] = at(x, y + 1);
			tv[1][0] = at(x + 1, y);
			tv[1][1] = at(x + 1, y + 1);
			unsigned int v[6][2] = {{x, y}, {x, y + 1}, {x + 1, y + 1}, {x + 1, y + 1}, {x + 1, y}, {x, y}};
			TerrainValue h[6];
			h[0] = tv[0][0];
			h[1] = tv[0][1];
			h[2] = tv[1][1];
			h[3] = tv[1][1];
			h[4] = tv[1][0];
			h[5] = tv[0][0];

			float texCoords[6][2] // [i][dimension]
			= {{1,0}, {1,1}, {0,1}, {0,1}, {0,0}, {1,0}};
			
			glBegin(GL_TRIANGLES);
			
			for (int i = 0; i < 6; i++) {
				glTexCoord2f(texCoords[i][0], texCoords[i][1]);
				glNormal3f(h[i].normal.x, h[i].normal.y, h[i].normal.z);
				glVertex3f(scale.x * float(v[i][0] - 1) / (terrainWidth - 3), scale.y * h[i].height, scale.z * float(v[i][1] - 1) / (terrainHeight - 3));
			}

			glEnd();
		}
	}

/*	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	for (x = 1; x < terrainWidth - 2; x++)
	{
		for (y = 1; y < terrainHeight - 2; y++)
		{
			glColor3f(1.0f, 0.0f, 0.0f);
			vertex normal = at(x, y).normal;
			glVertex3f(scale.x * (x - 1) / (terrainWidth - 3),
				scale.y * at(x, y).height,
				scale.z * (y - 1) / (terrainHeight - 3));
			glColor3f(0.0f, 0.0f, 1.0f);
			glVertex3f(scale.x * (x - 1) / (terrainWidth - 3) + normal.x,
				scale.y * at(x, y).height + normal.y,
				scale.z * (y - 1) / (terrainHeight - 3) + normal.z);
		}
	}
	glEnd();
	glEnable(GL_LIGHTING);*/
}

float Terrain::elevation(float x, float y) {
	if (x < 0.0f) x = 0.0f;
	if (y < 0.0f) y = 0.0f;
	if (x > terrainScale.x) x = terrainScale.x;
	if (y > terrainScale.z) y = terrainScale.z;
	int intX = int(x / terrainScale.x * (width() - 3) + 1);
	int intY = int(y / terrainScale.z * (height() - 3) + 1);
	if (intX < 0) intX = 0;
	if (intY < 0) intY = 0;
	if (intX + 1 >= (int)width()) intX = width() - 2;
	if (intY + 1 >= (int)height()) intY = height() - 2;
	float partX = (x - terrainScale.x * float(intX - 1) / (width() - 3)) / (terrainScale.x / (width() - 3));
	float partY = (y - terrainScale.z * float(intY - 1) / (height() - 3)) / (terrainScale.z / (height() - 3));
	float A = (1 - partX) * at(intX, intY).height + partX * at(intX + 1, intY).height;
	float B = (1 - partX) * at(intX, intY + 1).height + partX * at(intX + 1, intY + 1).height;
	float value = (1 - partY) * A + partY * B;
	return value * terrainScale.y;
}

vertex Terrain::normal(float x, float y) {
	if (x < 0.0f) x = 0.0f;
	if (y < 0.0f) y = 0.0f;
	if (x > terrainScale.x) x = terrainScale.x;
	if (y > terrainScale.z) y = terrainScale.z;
	int intX = int(x / terrainScale.x * (width() - 3) + 1);
	int intY = int(y / terrainScale.z * (height() - 3) + 1);
	if (intX < 0) intX = 0;
	if (intY < 0) intY = 0;
	if (intX + 1 >= (int)width()) intX = width() - 2;
	if (intY + 1 >= (int)height()) intY = height() - 2;
	float partX = (x - terrainScale.x * float(intX - 1) / (width() - 3)) / (terrainScale.x / (width() - 3));
	float partY = (y - terrainScale.z * float(intY - 1) / (height() - 3)) / (terrainScale.z / (height() - 3));
	vertex A = (1 - partX) * at(intX, intY).normal + partX * at(intX + 1, intY).normal;
	vertex B = (1 - partX) * at(intX, intY + 1).normal + partX * at(intX + 1, intY + 1).normal;
	return normalize((1 - partY) * A + partY * B);
}
